/*
 * File:   motorboard.c
 * Author: toffi
 *
 * Created on April 16, 2018, 11:52 PM
 */


#include <xc.h>
#include "pic_config.h"
#include "sw_config.h"
#include "led.h"
#include "pwm.h"
#include "can.h"
#include "qei.h"
#include "uart.h"
#include "state.h"

int main(void)
{
    // ========================================================================
    // Step 0: Oscillator Switch to get 140 MHz / 70 MIPS
    // No error handling here, if the oscillator fails we're fucked anyway
    // We have an 8MHz resonator, so we transform 8 -> 4 -> 280 -> 140 MHz
    PLLFBD = 68; // set PLL divider to 70 (from 4 to 280 MHz)
    CLKDIV = 0; // set N1 and N2 to 2 (8 to 4 and 280 to 140 MHz)
    __builtin_write_OSCCONH(0x03); // prepare clock switch to XTPLL
    __builtin_write_OSCCONL(OSCCON | 0x01); // initiate clock switch
    while(OSCCONbits.COSC != 0b011); // wait for XT to stabilize
    while(OSCCONbits.LOCK != 1); // wait for PLL lock
    // If we make it until here, the PIC is running at 70 MIPS.
    
    // ========================================================================
    // Step 1: I/O configuration
    // Normal Port Pin Configuration
    ANSELA = 0x0001; // AN0 is used for motor current measurement, all others off
    TRISA = 0xFFFF;
    TRISAbits.TRISA1 = 0; // RA1 is an output (LED red)
    
    ANSELB = 0; // no analog functions of port B used
    LATB = 0x0080; // ~CS of the encoder is high
    TRISB = 0xFFFF;
    TRISBbits.TRISB0 = 0; // RB0 (LED green)
    TRISBbits.TRISB1 = 0; // RB1 (LED blue)
    TRISBbits.TRISB7 = 0; // RB7 (Encoder ~CS
    TRISBbits.TRISB13 = 0; // RB13 (Motor PWM)
    TRISBbits.TRISB14 = 0; // RB14 (Motor direction)
    
    // Peripheral Pin Select Configuration
    // unlock PPS registers
    __builtin_write_OSCCONL(OSCCON & ~(1<<6));
    // CAN1 Module
    RPINR26bits.C1RXR = 36; // CAN Rx is mapped to RP36 / Pin RB4
    RPOR0bits.RP20R = 0x0E; // RP20 / Pin RA4 mapped to CAN Tx
    // QEI module
    RPINR14bits.QEA1R = 37; // assign pin RB5 / RP37 to QEI A 
    RPINR14bits.QEB1R = 42; // assign pin RB10 / RP42 to QEI B
    //RPINR15bits.INDX1R = 41; // assign pin RB9 / RP41 to INDEX
    RPINR0bits.INT1R = 41; // assign pin RB9 / RP41 to INT1 for homing
#if OWN_AXIS_NUMBER == 2
    // UART1
    RPINR18bits.U1RXR = 44; // Rx on RPI44 / RB12 -> J4
    RPOR4bits.RP43R = 1; // Tx on RP42 / RB11 -> J5
#endif
#ifdef CALIBRATION_MODE
    // UART1
    RPINR18bits.U1RXR = 44; // Rx on RPI44 / RB12 -> J4
    RPOR4bits.RP43R = 1; // Tx on RP42 / RB11 -> J5
#endif
    
    // lock PPS registers
    __builtin_write_OSCCONL(OSCCON | (1<<6));
    
    // ========================================================================
    // Step 2: PWM module configuration
    pwmInit(0xFFFF, 0); // for now, this is a really slow PWM at 1.8kHz, but resolution is great!
    
    // ========================================================================
    // Step 3: ADC configuration
    
    // ========================================================================
    // Step 4: Encoder interface configuration
    qeiInit();
    
    // ========================================================================
    // Step 5: CAN module configuration
    canInit();
    
    // ========================================================================
    // Step 5.1: UART configuration
#ifdef CALIBRATION_MODE
    uartInit();
#endif
    
    // ========================================================================
    // Step 6: System timer configuration
    // we use Timer2 in 16 bit mode
    T2CONbits.TCKPS0 = 1; // prescaler dividing by 8 -> counting @ 8750kHz
    PR2 = 8749; // counting from 0 to 8749 -> interrupts @ 1kHz
    IPC1bits.T2IP = 4; // lower priority than the UART
    IEC0bits.T2IE = 1; // enable timer 2 interrupt
    T2CONbits.TON = 1; // start timer
    
    // ========================================================================
    // Step 7: Interrupt configuration
    INTCON2bits.GIE = 1; // enable interrupts
    
#if OWN_AXIS_NUMBER == 2
    setLED(GREEN, ON);
#endif
    
    // ========================================================================
    // Step 8: Power saving... At least a little.
    PMD1bits.T5MD = 1;
    PMD1bits.T4MD = 1;
    PMD1bits.T3MD = 1;
    PMD1bits.T1MD = 1;
    PMD1bits.I2C1MD = 1;
    PMD1bits.U2MD = 1;
    PMD1bits.SPI2MD = 1;
    PMD1bits.SPI1MD = 1;
    PMD2 = 0xFFFF;
    PMD3bits.CMPMD = 1;
    PMD3bits.I2C2MD = 1;
    PMD4 = 0xFFFF;
    PMD6bits.PWM3MD = 1;
    PMD6bits.PWM1MD = 1;
    
    // This loop is a central part of the windows operating system:
    // it does nothing, but takes up any time available
    while(1);
    
    // We can never get here, but the compiler complains like a teenager if
    // main is not of type int, so it has to be int, and have a return...
    return(0);
}

// Timer 2 interrupt service routine
// is called every 1ms
void __attribute__((__interrupt__, __auto_psv__)) _T2Interrupt(void)
{
    
#ifdef CALIBRATION_MODE
    calibrationTimerInt();
#endif
    
    stateMsInterrupt();
    
    // end of timer 2 interrupt routine
    IFS0bits.T2IF = 0; // reset interrupt flag
}
