#include <xc.h>
#include "sw_config.h"
#include "pwm.h"
#include "qei.h"
#include "state.h"
#include "led.h"
#include "can.h"

// All state of this machine axis is in this structure. Call it an object if
// you believe in objects...
static struct
{
    unsigned int clockHours;
    unsigned int clockMinutes;
    int operatingMode;
    long targetPosition; // encoder value of our target position
    int targetSpeed; // the targeted maximum speed for a move
    int commandedSpeed; // the set speed, either by the position controller or
                        // by external command
    unsigned int positiveEndstop;
    unsigned int negativeEndstop;
} axisState =
{
    11, // Hours
    54, // Minutes
#ifdef MODULE_TEST_MODE
    10, // operatingMode
#else
    0, // operatingMode
#endif
    0, //targetPosition
    0, // targetSpeed
    0, // commandedSpeed
    0, // positiveEndstop
    0, // negativeEndstop
};

void stateAdvanceMinute(void);
void stateAdvanceHour(void);
void stateSendNewWheelPositions(void);
void stateMotorController(void);
#ifndef CALIBRATION_MODE
static unsigned int stateExpFilter(unsigned int oldValue, unsigned int newMeasurement, unsigned int m);
#endif

// Incoming speed commands are passed in here from the CAN receive interrupt,
// so this absolutely must stay super short
void stateIncomingSpeed(int speed)
{
    setLED(RED, ON);
    long positionTemp;
    // Now, for use in the clock, this routine is different to what it will
    // later be, as the "speed" command is interpreted as a "target number to
    // show" command ranging from 0 to 9, with negative numbers -1 to -10 used
    // to have the wheel turn backwards
    if(speed >= 0) // command is forward to a number
    {
        if(speed < 10) // only digits 0 to 9 are valid
        {
            if((speed * 1638) < axisState.targetPosition) // a rollover will occur
            {
                positionTemp = qeiReadPosition();
                positionTemp -= 16383;
                qeiWritePosition(positionTemp);
                axisState.targetSpeed = CLOCK_MOVE_SPEED;
                axisState.targetPosition = speed * (long) 1638;
            }
            else if((speed * 1638) > axisState.targetPosition) // normal move
            {
                axisState.targetSpeed = CLOCK_MOVE_SPEED;
                axisState.targetPosition = speed * (long) 1638;
            }
        }
    }
    else // command is backwards to a number
    {
        if(speed > -11) // only -1 to -10 are valid
        {
            if((speed * 1638) >= axisState.targetPosition) // a rollover will occur
            {
                positionTemp = qeiReadPosition();
                positionTemp += 16383;
                qeiWritePosition(positionTemp);
            }
            axisState.targetSpeed = CLOCK_MOVE_SPEED * (-1);
            axisState.targetPosition = ((speed * (-1)) + 1) * (long) 1638;
        }
    }
}

void stateIncomingButtons(unsigned int buttons)
{
    switch(buttons)
    {
        case 0: // minutes button
            stateAdvanceMinute();
            break;
        case 1:
            axisState.clockMinutes += 10;
            if(axisState.clockMinutes >= 60) // hour rollover
            {
                axisState.clockMinutes = 0;
            }
            stateSendNewWheelPositions();
            break;
        case 3:
            stateAdvanceHour();
            stateSendNewWheelPositions();
            break;
    }
}

void stateMsInterrupt(void)
{
    static unsigned int msCounter = 0, tenCounter = 0, hundretCounter = 0, secondCounter = 0;
#ifndef CALIBRATION_MODE
    unsigned int tempEndstop; // used for debouncing endstop Inputs
#if OWN_AXIS_NUMBER != 4
    CAN_MESSAGE_T canMessageOut;
#endif
#endif
    
    // advance counters
    msCounter++;
    if(msCounter >= 10) // 10 ms have accumulated
    {
        msCounter = 0;
        tenCounter++;
        if(tenCounter >= 10)
        {
            tenCounter = 0;
            hundretCounter++;
            if(hundretCounter >= 10)
            {
                hundretCounter = 0;
                secondCounter++;
                if(secondCounter >= 60)
                {
                    secondCounter = 0;
                    // any stuff to happen every minute goes in below here:
#if OWN_AXIS_NUMBER == 4
                    stateAdvanceMinute();
#endif
                }
                // any stuff to happen every second goes in below here:
            }
            // any stuff to happen every 100 ms goes in below here:
        }
        // any stuff to happen every 10 ms goes in below here:
    }
    // any stuff to happen every 1 ms goes in below here:
#ifndef CALIBRATION_MODE
#if OWN_AXIS_NUMBER != 2
    // movement control for the axis lives in here:
    switch(axisState.operatingMode)
    {
        case 0: // uninitialised, start homing
            qeiSetSimpleHomingMode();
            pwmSetDirection(FORWARD);
            pwmSetDutyCycle(CLOCK_HOME_SPEED);
            setLED(RED, ON);
            axisState.operatingMode = 1;
            break;
        case 1: // waiting for index / zeroing, then stop
            if(qeiCheckHomingMode() == 0) // homing has finished
            {
                pwmSetDutyCycle(0);
                setLED(YELLOW, ON);
                axisState.operatingMode = 2;
            }
            break;
        case 2: // start driving to display "0"
            qeiWritePosition(ZERO_OFFSET);
            pwmSetDirection(FORWARD);
            pwmSetDutyCycle(CLOCK_MOVE_SPEED);
            axisState.operatingMode = 3;
            break;
        case 3: // stop at "0"
            if(qeiReadPosition() > 0)
            {
                pwmSetDutyCycle(0);
                setLED(WHITE, OFF);
                axisState.operatingMode = 10;
            }
            break;
        case 10: // normal operation
            stateMotorController();
            break;
        default:
            axisState.operatingMode = 0;
    }
#endif // if not module 2
    // button reading lives in here:
    // debounce button first
    if(!PORTBbits.RB12) // button connected to J4 reads as pressed
    {
        tempEndstop = stateExpFilter(axisState.positiveEndstop, 100, 3);
    }
    else // button reads as up
    {
        tempEndstop = stateExpFilter(axisState.positiveEndstop, 0, 3);
    }
    // check for change of state:
    if(tempEndstop > 70 && axisState.positiveEndstop < 70) // unpressed to pressed transition
    {
        setLED(BLUE, ON);
#if OWN_AXIS_NUMBER != 4
        // if we are one of the "normal" modules, we send out the button input
        // to module 4 to change time
        canMessageOut.IDE = 0; // no extended addressing
        canMessageOut.SRR = 0; // no remote frame
        canMessageOut.SID = 0x0002; // clock sync... bastardised to send out inputs
        canMessageOut.DLC = 2; // 2 Bytes of data
        canMessageOut.unused_null = 0; // there are two reserved 0-bits in there
        canMessageOut.RTR = 0; // not a remote frame
        canMessageOut.WORD0 = (unsigned int) OWN_AXIS_NUMBER;
        canSendMessage(1, &canMessageOut);
#endif
#if OWN_AXIS_NUMBER == 4
        // if we are module 4, we change time directly
        // pressing this button advances time by 10 hours, with a wraparound highly likely
        axisState.clockHours += 10;
        if(axisState.clockHours >= 24) // wraparound...
        {
            axisState.clockHours -= 20; // we only reset the ten hours wheel
        }
        // now, set the wheels...
        stateIncomingSpeed(axisState.clockHours / 10);
#endif
    }
    else if(tempEndstop <30) // pressed to unpressed transition
    {
        setLED(BLUE, OFF);
    }
    axisState.positiveEndstop = tempEndstop;
#endif // if not in calibration mode...
}

void stateAdvanceMinute(void)
{
    axisState.clockMinutes++;
    if(axisState.clockMinutes >= 60) // hour rollover
    {
        axisState.clockMinutes = 0;
        stateAdvanceHour();
    }
    stateSendNewWheelPositions();
}

void stateAdvanceHour(void)
{
    axisState.clockHours++;
    if(axisState.clockHours >= 24) // day rollover
    {
        axisState.clockHours = 0;
    }
}

void stateSendNewWheelPositions(void)
{
    int wheel0, wheel1, wheel3, wheel4; // temporary storage for wheel positions
    CAN_MESSAGE_T canMessageOut;
    
    // first, generate wheel positions from time:
    wheel0 = axisState.clockMinutes % 10;
    wheel1 = axisState.clockMinutes / 10; // This is actually a floor division!
    wheel3 = axisState.clockHours % 10;
    wheel4 = axisState.clockHours / 10; // Again, this is floor, not division
    
    // then, set those wheel positions:
    // CAN message for wheel 0, 1 and 3:
    canMessageOut.IDE = 0; // no extended addressing
    canMessageOut.SRR = 0; // no remote frame
    canMessageOut.SID = 0x0010; // axis control, axis 0-3
    canMessageOut.DLC = 8; // 8 Bytes of data
    canMessageOut.unused_null = 0; // there are two reserved 0-bits in there
    canMessageOut.RTR = 0; // not a remote frame
    canMessageOut.WORD0 = wheel0;
    canMessageOut.WORD1 = wheel1;
    canMessageOut.WORD2 = 0;
    canMessageOut.WORD3 = wheel3;
    canSendMessage(0, &canMessageOut);
    // direct setting of wheel 4:
    stateIncomingSpeed(wheel4);
}

void stateMotorController(void) // shit simple three-point controller
{
    if(axisState.targetSpeed >= 0) // going forward
    {
        if(qeiReadPosition() < axisState.targetPosition) // not at target
        {
            pwmSetDirection(FORWARD);
            pwmSetDutyCycle(CLOCK_MOVE_SPEED);
            setLED(GREEN, OFF);
            setLED(RED, ON);
        }
        else // target reached / overshot -> stop
        {
            pwmSetDutyCycle(0);
            setLED(RED, OFF);
            setLED(GREEN, ON);
        }
    }
    else if(axisState.targetSpeed < 0) // going backwards
    {
        if(qeiReadPosition() > axisState.targetPosition) // not at target
        {
            pwmSetDirection(REVERSE);
            pwmSetDutyCycle(CLOCK_MOVE_SPEED);
            setLED(GREEN, OFF);
            setLED(RED, ON);
        }
        else // target reached / overshot -> stop
        {
            pwmSetDutyCycle(0);
            setLED(RED, OFF);
            setLED(GREEN, ON);
        }
    }
}

#ifndef CALIBRATION_MODE
// A simple 1st order IIR filter for 16 bit (unsigned) integers
// It follows the basic filter equation of y_n = (k-1)x_n + ky_n-1, but arranged
// slightly different for better numeric performance. The filter coefficient "m"
// can be derived as m = 1/k - 1.
// The actual implementation is y_n = (x_n + my_n-1) / m+1
static unsigned int stateExpFilter(unsigned int oldValue, unsigned int newMeasurement, unsigned int m)
/*{
    unsigned long temp;
    temp = (unsigned long) m * (unsigned long) oldValue + (unsigned long) newMeasurement;
    temp /= m+1;
    return (unsigned int) temp;
}*/
{ // m = 3
    unsigned long temp;
    temp = (unsigned long) 3 * (unsigned long) oldValue + (unsigned long) newMeasurement;
    temp = temp >> 2;
    return (unsigned int) temp;
}
#endif
